from dataclasses import dataclass

from waffle_hub.schema.base_schema import BaseSchema


@dataclass
class ObjectDetectionMetric(BaseSchema):
    mAP: float
    mAP_50: float
    mAP_75: float
    mAP_small: float
    mAP_medium: float
    mAP_large: float
    mAR_1: float
    mAR_10: float
    mAR_100: float
    mAR_small: float
    mAR_medium: float
    mAR_large: float
    mAP_per_class: list[float]
    mAR_100_per_class: list[float]


@dataclass
class ClassificationMetric(BaseSchema):
    accuracy: float
    recall: float
    precision: float
    f1_score: float

    accuracy_per_class: list[float]
    recall_per_class: list[float]
    precision_per_class: list[float]
    f1_score_per_class: list[float]

    confusion_matrix: list[list[int]]


@dataclass
class InstanceSegmentationMetric(BaseSchema):
    mAP: float


@dataclass
class TextRecognitionMetric(BaseSchema):
    accuracy: float


@dataclass
class SemanticSegmentationMetric(BaseSchema):
    mean_pixel_accuracy: float
    IoU: float



from pathlib import Path

from tqdm import tqdm
from waffle_utils.file import io
from waffle_utils.image.io import load_image

from waffle_hub import TaskType
from waffle_hub.schema.fields import Annotation, Category, Image


def import_object_detection(self, json_file, image_dir=None):
    """
    Label studio object detection format

    [
        {
            "id": 1,
            "data": {
                "image": "/data/upload/1/1.jpg"  # {data_path}/{project_id}/{file_name}
            },
            "file_upload": "1.jpg",  # {file_name}
            "annotations": [
                {
                    "id": 1,
                    "result": [
                        {
                            "id": 1,
                            "type": "rectanglelabels",
                            "value": {
                                "rectanglelabels": ["cat"],
                                "x": 20.3,  # left  # 0 ~ 100 (%)
                                "y": 34.2,  # top  # 0 ~ 100 (%)
                                "width": 10.4,  # width  # 0 ~ 100 (%)
                                "height": 10.2  # height  # 0 ~ 100 (%)
                            },
                            ...
                        },
                        ...
                    ],
                    ...
                },
                ...
            ],
            ...
        },
        ...
    ]
    """

    images = []
    annotations = []
    categories = []
    category_to_id = {}

    datas = io.load_json(json_file)

    for image_id, data in tqdm(enumerate(datas, start=1), total=len(datas)):

        image_file_name = data["file_upload"]
        src_image_path = Path(image_dir) / image_file_name if image_dir else data["data"]["image"]
        io.copy_file(src_image_path, self.raw_image_dir / image_file_name, create_directory=True)

        H, W = load_image(src_image_path).shape[:2]

        image = Image.new(
            image_id=image_id,
            file_name=str(image_file_name),
            width=W,
            height=H,
        )
        images.append(image)

        for annotation in data["annotations"]:
            for result in annotation["result"]:
                if "rectanglelabels" not in result["value"]:
                    continue
                value = result["value"]

                category = value["rectanglelabels"][0]
                if category not in category_to_id:
                    category_to_id[category] = len(category_to_id) + 1
                    categories.append(
                        Category.object_detection(
                            category_id=category_to_id[category],
                            name=category,
                            supercategory="object",
                        )
                    )

                x = value["x"] * W / 100
                y = value["y"] * H / 100
                width = value["width"] * W / 100
                height = value["height"] * H / 100

                annotations.append(
                    Annotation.object_detection(
                        annotation_id=len(annotations) + 1,
                        image_id=image_id,
                        category_id=category_to_id[category],
                        bbox=[x, y, width, height],
                        is_crowd=False,
                    )
                )

    self.add_images(images)
    self.add_categories(categories)
    self.add_annotations(annotations)


def import_classification(self, json_file, image_dir):
    """
    Label studio classification format

    [
        {
            "id": 1,
            "data": {
                "image": "/data/upload/1/1.jpg"  # {data_path}/{project_id}/{file_name}
            },
            "file_upload": "1.jpg",  # {file_name}
            "annotations": [
                {
                    "id": 1,
                    "result": [
                        {
                            "id": 1,
                            "type": "choices",
                            "value": {
                                "choices": ["cat"]
                            },
                            ...
                        },
                        ...
                    ],
                    ...
                },
                ...
            ],
            ...
        },
        ...
    ]
    """

    images = []
    annotations = []
    categories = []
    category_to_id = {}

    datas = io.load_json(json_file)

    for image_id, data in tqdm(enumerate(datas, start=1), total=len(datas)):

        image_file_name = data["file_upload"]
        src_image_path = Path(image_dir) / image_file_name if image_dir else data["data"]["image"]
        io.copy_file(src_image_path, self.raw_image_dir / image_file_name, create_directory=True)

        H, W = load_image(src_image_path).shape[:2]

        image = Image.new(
            image_id=image_id,
            file_name=str(image_file_name),
            width=W,
            height=H,
        )
        images.append(image)

        for annotation in data["annotations"]:
            for result in annotation["result"]:
                if "choices" not in result["value"]:
                    continue
                value = result["value"]

                category = value["choices"][0]
                if category not in category_to_id:
                    category_to_id[category] = len(category_to_id) + 1
                    categories.append(
                        Category.classification(
                            category_id=category_to_id[category],
                            name=category,
                        )
                    )

                annotations.append(
                    Annotation.classification(
                        annotation_id=len(annotations) + 1,
                        image_id=image_id,
                        category_id=category_to_id[category],
                    )
                )

    self.add_images(images)
    self.add_categories(categories)
    self.add_annotations(annotations)


def import_label_studio(self, json_file, task, image_dir=None):

    if task == TaskType.OBJECT_DETECTION:
        _import = import_object_detection
    elif task == TaskType.CLASSIFICATION:
        _import = import_classification
    else:
        raise NotImplementedError(f"Task type {task} not implemented")

    _import(self, json_file, image_dir)

    # TODO: add unlabeled set
    io.save_json([], self.unlabeled_set_file, create_directory=True)



import torch


def device_context(func):
    def wrapper(*args, **kwargs):
        if torch.cuda.is_available():
            if hasattr(torch._C, "_cuda_clearCublasWorkspaces"):  # for torch 2.0 versions
                # https://github.com/pytorch/pytorch/issues/95668
                torch._C._cuda_clearCublasWorkspaces()
            # Memory free
            torch.cuda.empty_cache()
        result = func(*args, **kwargs)
        if torch.cuda.is_available():
            if hasattr(torch._C, "_cuda_clearCublasWorkspaces"):
                # https://github.com/pytorch/pytorch/issues/95668
                torch._C._cuda_clearCublasWorkspaces()
            torch.cuda.empty_cache()
        return result

    return wrapper



from .configs import ExportOnnxConfig, InferenceConfig, ModelConfig, TrainConfig
from .data import DatasetInfo, ImageInfo
from .fields import Annotation, Category, Image

__all__ = [
    "ModelConfig",
    "TrainConfig",
    "InferenceConfig",
    "ExportOnnxConfig",
    "DatasetInfo",
    "ImageInfo",
    "Image",
    "Category",
    "Annotation",
]



from itertools import permutations

import numpy as np
import pytest

from waffle_hub.schema.evaluate import (
    ClassificationMetric,
    InstanceSegmentationMetric,
    SemanticSegmentationMetric,
    ObjectDetectionMetric,
)
from waffle_hub.schema.fields import Annotation
from waffle_hub.utils.data import resize_image
from waffle_hub.utils.evaluate import (
    evaluate_classification,
    evaluate_instance_segmentation,
    evaluate_object_detection,
    evalute_semantic_segmentation
)


def test_evaluate_classification():
    result: ClassificationMetric = evaluate_classification(
        preds=[
            [Annotation.classification(category_id=1)],
            [Annotation.classification(category_id=2)],
            [Annotation.classification(category_id=1)],
        ],
        labels=[
            [Annotation.classification(category_id=1)],
            [Annotation.classification(category_id=2)],
            [Annotation.classification(category_id=3)],
        ],
        num_classes=3,
    )

    assert abs(result.accuracy - 2 / 3) < 1e-5


def test_evaluate_object_detection():
    result: ObjectDetectionMetric = evaluate_object_detection(
        preds=[
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=3, bbox=[0, 0, 1, 1], score=1.0),
            ],
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
            ],
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
            ],
        ],
        labels=[
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=3, bbox=[0, 0, 1, 1], score=1.0),
            ],
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=3, bbox=[0, 0, 1, 1], score=1.0),
            ],
            [
                Annotation.object_detection(category_id=1, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=2, bbox=[0, 0, 1, 1], score=1.0),
                Annotation.object_detection(category_id=3, bbox=[0, 0, 1, 1], score=1.0),
            ],
        ],
        num_classes=3,
    )

    assert result.mAP < 1.0


def test_evaluate_instance_segmentation():
    result: InstanceSegmentationMetric = evaluate_instance_segmentation(
    preds=[
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.instance_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.instance_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.instance_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
        ],
        labels=[
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
            [
                Annotation.instance_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.instance_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
        ],
        num_classes=3,
        image_size=[32,32],
    )

    assert result.mAP < 1.0 and result.mAP > 0.0


def test_evalute_semantic_segmentation():
    result: SemanticSegmentationMetric = evalute_semantic_segmentation(
    preds=[
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.semantic_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.semantic_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
                Annotation.semantic_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]], score=1.0),
            ],
        ],
        labels=[
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
            [
                Annotation.semantic_segmentation(category_id=1, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=2, segmentation=[[0, 0, 1, 0, 1, 1]]),
                Annotation.semantic_segmentation(category_id=3, segmentation=[[0, 0, 1, 0, 1, 1]]),
            ],
        ],
        num_classes=3,
        image_size=[32,32],
    )

    assert result.IoU < 1.0 and result.IoU > 0.0
    assert result.mean_pixel_accuracy < 1.0 and result.mean_pixel_accuracy > 0.0

def test_resize_function():
    letter_box = [False, True]
    image_size = [224, 480, 512, 640, 1280, 1600, 1920]

    rect_list = permutations(image_size, 2)
    square_list = [(sz, sz) for sz in image_size]

    for lb in letter_box:
        # square 2 square
        for ori_shape in square_list:
            for resize_shape in square_list:
                image, info = resize_image(
                    np.random.randint(0, 255, ori_shape + (3,), dtype=np.uint8),
                    resize_shape,
                    letter_box=lb,
                )
                assert (image.shape[1], image.shape[0]) == info[
                    "new_shape"
                ], f"Difference from resize image shape to info (letter_box: {lb})"
                if lb:
                    assert (
                        (info["input_shape"][0] - info["new_shape"][0]) // 2,
                        (info["input_shape"][1] - info["new_shape"][1]) // 2,
                    ) == info["pad"], f"Difference Pad info (letter_box: {lb})"
                else:
                    assert (
                        info["input_shape"] == info["new_shape"]
                    ), f"Difference from info[input_shape] to info[new_shape] (letter_box: {lb})"
        # square 2 rect
        for ori_shape in square_list:
            for resize_shape in rect_list:
                image, info = resize_image(
                    np.random.randint(0, 255, ori_shape + (3,), dtype=np.uint8),
                    resize_shape,
                    letter_box=lb,
                )
                assert (image.shape[1], image.shape[0]) == info[
                    "new_shape"
                ], f"Difference from resize image shape to info (letter_box: {lb})"
                if lb:
                    assert (
                        (info["input_shape"][0] - info["new_shape"][0]) // 2,
                        (info["input_shape"][1] - info["new_shape"][1]) // 2,
                    ) == info["pad"], f"Difference Pad info (letter_box: {lb})"
                else:
                    assert (
                        info["input_shape"] == info["new_shape"]
                    ), f"Difference from info[input_shape] to info[new_shape] (letter_box: {lb})"
        # rect 2 square
        for ori_shape in rect_list:
            for resize_shape in square_list:
                image, info = resize_image(
                    np.random.randint(0, 255, ori_shape + (3,), dtype=np.uint8),
                    resize_shape,
                    letter_box=lb,
                )
                assert (image.shape[1], image.shape[0]) == info[
                    "new_shape"
                ], f"Difference from resize image shape to info (letter_box: {lb})"
                if lb:
                    assert (
                        (info["input_shape"][0] - info["new_shape"][0]) // 2,
                        (info["input_shape"][1] - info["new_shape"][1]) // 2,
                    ) == info["pad"], f"Difference Pad info (letter_box: {lb})"
                else:
                    assert (
                        info["input_shape"] == info["new_shape"]
                    ), f"Difference from info[input_shape] to info[new_shape] (letter_box: {lb})"
        # rect 2 rect
        for ori_shape in rect_list:
            for resize_shape in rect_list:
                image, info = resize_image(
                    np.random.randint(0, 255, ori_shape + (3,), dtype=np.uint8),
                    resize_shape,
                    letter_box=lb,
                )
                assert (image.shape[1], image.shape[0]) == info[
                    "new_shape"
                ], f"Difference from resize image shape to info (letter_box: {lb})"
                if lb:
                    assert (
                        (info["input_shape"][0] - info["new_shape"][0]) // 2,
                        (info["input_shape"][1] - info["new_shape"][1]) // 2,
                    ) == info["pad"], f"Difference Pad info (letter_box: {lb})"
                else:
                    assert (
                        info["input_shape"] == info["new_shape"]
                    ), f"Difference from info[input_shape] to info[new_shape] (letter_box: {lb})"



import json
import tempfile
import time
from pathlib import Path

import pytest
import torch

from waffle_hub import (
    EvaluateStatus,
    ExportStatus,
    InferenceStatus,
    TaskType,
    TrainStatus,
)
from waffle_hub.dataset import Dataset
from waffle_hub.hub import Hub
from waffle_hub.schema.result import (
    EvaluateResult,
    ExportOnnxResult,
    ExportWaffleResult,
    InferenceResult,
    TrainResult,
)


def _train(hub, dataset: Dataset, image_size: int, advance_params: dict = None):
    result: TrainResult = hub.train(
        dataset=dataset,
        epochs=1,
        image_size=image_size,
        batch_size=4,
        pretrained_model=None,
        letter_box=False,
        device="cpu",
        workers=0,
        advance_params=advance_params,
    )

    training_info = hub.get_training_info()
    assert training_info.status == TrainStatus.SUCCESS
    assert training_info.step == training_info.total_step
    assert len(result.metrics) >= 1
    assert len(result.eval_metrics) >= 1
    assert Path(result.best_ckpt_file).exists()
    # assert Path(result.last_ckpt_file).exists()

    if hub.backend == "ultralytics":
        if hub.task in [TaskType.INSTANCE_SEGMENTATION, TaskType.OBJECT_DETECTION]:
            assert hub.get_train_config().letter_box == True
        elif hub.task == TaskType.CLASSIFICATION:
            assert hub.get_train_config().letter_box == False

    return result


def _evaluate(hub, dataset: Dataset):

    result: EvaluateResult = hub.evaluate(
        dataset=dataset,
        device="cpu",
        workers=0,
    )

    evaluating_info = hub.get_evaluating_info()
    assert evaluating_info.status == EvaluateStatus.SUCCESS
    assert evaluating_info.step == evaluating_info.total_step
    assert len(result.eval_metrics) >= 1

    return result


def _inference(hub, source: str):

    result: InferenceResult = hub.inference(
        source=source,
        draw=True,
        device="cpu",
        workers=0,
    )

    inferencing_info = hub.get_inferencing_info()
    assert inferencing_info.status == InferenceStatus.SUCCESS
    assert inferencing_info.step == inferencing_info.total_step
    assert len(result.predictions) >= 1
    assert Path(result.draw_dir).exists()

    return result


def _export_onnx(hub, half: bool = False):
    result: ExportOnnxResult = hub.export_onnx(
        half=half,
        device="cpu",
    )

    exporting_onnx_info = hub.get_exporting_onnx_info()
    assert exporting_onnx_info.status == ExportStatus.SUCCESS
    assert Path(result.onnx_file).exists()

    return result


def _export_waffle(hub):
    result: ExportWaffleResult = hub.export_waffle()

    assert Path(result.waffle_file).exists()

    return result


def _from_waffle_file(waffle_file: str, source: str, tmpdir: Path):
    name = "test_import"
    hub = Hub.from_waffle_file(name=name, waffle_file=waffle_file, root_dir=tmpdir)

    _inference(hub, source)


def _feature_extraction(
    hub,
    image_size: int,
):

    model = hub.get_model()

    layer_names = model.get_layer_names()
    assert len(layer_names) > 0

    x = torch.randn(1, 3, image_size, image_size)
    layer_name = layer_names[-1]
    x, feature_maps = model.get_feature_maps(x, layer_name)
    assert len(feature_maps) == 1


def _benchmark(hub, image_size):
    hub.benchmark(device="cpu", half=False, image_size=image_size)


def _util(hub):
    name = hub.name
    backend = hub.backend
    root_dir = hub.root_dir

    hub_class = Hub.get_hub_class(backend)
    assert hub_class == type(hub)

    hub_loaded = Hub.load(name, root_dir)
    assert isinstance(hub_loaded, type(hub))


def _total(
    hub,
    dataset: Dataset,
    image_size: int,
    tmpdir: Path,
    advance_params: dict = None,
):

    _train(hub, dataset, image_size, advance_params=advance_params)
    _evaluate(hub, dataset)
    _inference(hub, dataset.raw_image_dir)
    _export_onnx(hub, half=False)  # _export_onnx(hub, half=True)  # cpu cannot be half
    result = _export_waffle(hub)
    _from_waffle_file(result.waffle_file, dataset.raw_image_dir, tmpdir)
    _feature_extraction(hub, image_size)
    _benchmark(hub, image_size)
    _util(hub)


def test_ultralytics_segmentation(
    instance_segmentation_dataset: Dataset, tmpdir: Path, test_video_path: Path
):
    image_size = 32
    dataset = instance_segmentation_dataset

    # test hub
    name = "test_seg"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.INSTANCE_SEGMENTATION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)
    _inference(hub, test_video_path)


def test_ultralytics_object_detection(
    object_detection_dataset: Dataset, tmpdir: Path, test_video_path: Path
):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.OBJECT_DETECTION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)
    _inference(hub, test_video_path)


def test_ultralytics_object_detection_advance_params(
    object_detection_dataset: Dataset, tmpdir: Path
):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det_adv"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.OBJECT_DETECTION,
        model_type="yolov8",
        model_size="n",
        categories=dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    hub.get_default_advance_train_params()

    import_hub_name = "test_import"
    _total(hub, dataset, image_size, tmpdir, {"box": 4, "cls": 1})
    hub.delete_artifact()
    import_hub = Hub.load(name=import_hub_name, root_dir=tmpdir)
    import_hub.delete_hub()

    with open(str(tmpdir / "adv.json"), "w") as f:
        json.dump({"box": 4, "cls": 2}, f)
    _total(hub, dataset, image_size, tmpdir, str(tmpdir / "adv.json"))
    hub.delete_artifact()
    import_hub = Hub.load(name=import_hub_name, root_dir=tmpdir)
    import_hub.delete_hub()

    with pytest.raises(ValueError):
        _total(hub, dataset, image_size, tmpdir, {"box": 4, "dummy_adv_param": 2})
        assert hub.get_training_info().status == TrainStatus.FAILED
        import_hub = Hub.load(name=import_hub_name, root_dir=tmpdir)
        import_hub.delete_hub()


def test_ultralytics_classification(
    classification_dataset: Dataset, tmpdir: Path, test_video_path: Path
):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.CLASSIFICATION,
        model_type="yolov8",
        model_size="n",
        categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)
    _inference(hub, test_video_path)


def test_transformers_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="transformers",
        task=TaskType.OBJECT_DETECTION,
        model_type="YOLOS",
        model_size="tiny",
        categories=object_detection_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_autocare_dlt_object_detection(object_detection_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = object_detection_dataset

    # test hub
    name = "test_det"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.OBJECT_DETECTION,
        model_type="YOLOv5",
        model_size="s",
        categories=object_detection_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_autocare_dlt_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.CLASSIFICATION,
        model_type="Classifier",
        model_size="s",
        categories=dataset.get_categories(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_autocare_dlt_text_recognition(text_recognition_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = text_recognition_dataset

    # test hub
    name = "test_ocr"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.TEXT_RECOGNITION,
        model_type="TextRecognition",
        model_size="s",
        categories=dataset.get_categories(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_autocare_dlt_semantic_segmentation(semantic_segmentation_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = semantic_segmentation_dataset

    # test hub
    name = "test_segmantic_seg"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.SEMANTIC_SEGMENTATION,
        model_type="Segmenter",
        model_size="m",
        categories=semantic_segmentation_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_autocare_dlt_classification_without_category(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="autocare_dlt",
        task=TaskType.CLASSIFICATION,
        model_type="Classifier",
        model_size="s",
        # categories=dataset.get_categories(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_ultralytics_classification_without_category(classification_dataset: Dataset, tmpdir: Path):
    image_size = 32
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="ultralytics",
        task=TaskType.CLASSIFICATION,
        model_type="yolov8",
        model_size="n",
        # categories=classification_dataset.get_category_names(),  # auto detect
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)


def test_transformers_classification(classification_dataset: Dataset, tmpdir: Path):
    image_size = 224
    dataset = classification_dataset

    # test hub
    name = "test_cls"
    hub = Hub.new(
        name=name,
        backend="transformers",
        task=TaskType.CLASSIFICATION,
        model_type="ViT",
        model_size="tiny",
        # categories=classification_dataset.get_category_names(),
        root_dir=tmpdir,
    )
    hub = Hub.load(name=name, root_dir=tmpdir)
    hub: Hub = Hub.from_model_config(
        name=name + "_from_model_config",
        model_config_file=tmpdir / name / Hub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )

    _total(hub, dataset, image_size, tmpdir)



import logging
from pathlib import Path
from typing import Union

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from waffle_utils.file.network import get_file_from_url
from waffle_utils.image.io import load_image

from waffle_hub import TaskType
from waffle_hub.schema.fields import Annotation

FONT_URL = "https://raw.githubusercontent.com/snuailab/assets/main/waffle/fonts/gulim.ttc"
FONT_NAME = "gulim.ttc"


# random colors with 1000 categories
colors = np.random.randint(0, 255, (1000, 3), dtype="uint8").tolist()


def draw_classification(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    loc_x: int = 10,
    loc_y: int = 30,
):
    category_id: int = annotation.category_id
    score: float = annotation.score

    # calculate font size and thickness
    font_scale = max(image.shape[0], image.shape[1]) / 1000
    font_scale = 1.0 if font_scale < 1.0 else font_scale
    font_size = int(font_scale * 25)
    thinckness = int(font_scale) * 2

    # download font
    try:
        if not Path(FONT_NAME).exists():
            get_file_from_url(FONT_URL, FONT_NAME, True)

        font = ImageFont.truetype(FONT_NAME, font_size)
    except:
        font = ImageFont.load_default()
        logging.warning("Don't load font file, Using default font.")

    img_pil = Image.fromarray(image)
    draw = ImageDraw.Draw(img_pil)
    draw.text(
        (loc_x, loc_y - font_size),
        f"{names[category_id-1]}" + (f": {score:.2f}" if score else ""),
        font=font,
        fill=tuple(colors[category_id - 1]),
        stroke_width=thinckness,
    )
    image = np.array(img_pil)

    return image


def draw_object_detection(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    score: float = None,
):
    x1, y1, w, h = annotation.bbox
    x2 = x1 + w
    y2 = y1 + h

    category_id: int = annotation.category_id
    score: float = annotation.score

    # calculate font size and thickness
    font_scale = max(image.shape[0], image.shape[1]) / 1000
    font_scale = 1.0 if font_scale < 1.0 else font_scale
    font_size = int(font_scale * 15)
    thinckness = int(font_scale) * 2

    # download font
    try:
        if not Path(FONT_NAME).exists():
            get_file_from_url(FONT_URL, FONT_NAME, True)
        font = ImageFont.truetype(FONT_NAME, font_size)
    except:
        font = ImageFont.load_default()
        logging.warning("Don't load font file, Using default font.")

    img_pil = Image.fromarray(image)
    draw = ImageDraw.Draw(img_pil)
    draw.text(
        (int(x1), int(y1) - font_size),
        f"{names[category_id-1]}" + (f": {score:.2f}" if score else ""),
        font=font,
        fill=tuple(colors[category_id - 1]),
        stroke_width=thinckness,
    )

    draw.rectangle(
        (int(x1), int(y1), int(x2), int(y2)),
        outline=tuple(colors[category_id - 1]),
        width=thinckness,
    )

    image = np.array(img_pil)

    return image


def draw_instance_segmentation(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
    score: float = None,
):
    image = draw_object_detection(image, annotation, names, score)
    segments: list = annotation.segmentation

    if len(segments) == 0:
        return image

    pil_image = Image.fromarray(image)
    draw = ImageDraw.Draw(pil_image, "RGBA")
    fill_color = tuple(colors[annotation.category_id - 1])
    fill_color = fill_color + (120,)
    for segment in segments:
        draw.polygon(
            segment,
            fill=fill_color,
        )

    image = np.array(pil_image)

    return image


def draw_text_recognition(
    image: np.ndarray,
    annotation: Annotation,
    loc_x: int = 0,
    loc_y: int = 10,
):
    # calculate font size and thickness
    font_scale = max(image.shape[0], image.shape[1]) / 1000
    font_size = int((0.7 if font_scale < 0.7 else font_scale) * 25)
    thinckness = int(font_scale) * 2

    # download font
    try:
        if not Path(FONT_NAME).exists():
            get_file_from_url(FONT_URL, FONT_NAME, True)
        font = ImageFont.truetype(FONT_NAME, font_size)
    except:
        font = ImageFont.load_default()
        logging.warning("Don't load font file, Using default font.")

    img_pil = Image.fromarray(image)
    draw = ImageDraw.Draw(img_pil)
    draw.text(
        (loc_x, loc_y),
        annotation.caption,
        font=font,
        fill=tuple(colors[0]),
        stroke_width=thinckness,
    )

    image = np.array(img_pil)

    return image


def draw_semantic_segmentation(
    image: np.ndarray,
    annotation: Annotation,
    names: list[str],
):
    segments: list = annotation.segmentation

    if len(segments) == 0:
        return image

    pil_image = Image.fromarray(image)
    draw = ImageDraw.Draw(pil_image, "RGBA")
    fill_color = tuple(colors[annotation.category_id - 1])
    fill_color = fill_color + (120,)
    for segment in segments:
        if len(segment) < 6:
            continue

        draw.polygon(
            segment,
            fill=fill_color,
        )

    image = np.array(pil_image)

    return image


def draw_results(
    image: Union[np.ndarray, str],
    results: list[Annotation],
    names: list[str],
):

    if isinstance(image, str):
        image = load_image(image)

    task_results = {task: [] for task in TaskType}
    for result in results:
        task_results[result.task.upper()].append(result)

    font_scale = max(image.shape[0], image.shape[1]) / 1000
    font_scale = 1.0 if font_scale < 1.0 else font_scale
    font_size = int(font_scale * 25)
    for i, result in enumerate(task_results[TaskType.CLASSIFICATION], start=1):
        image = draw_classification(
            image,
            result,
            names=names,
            loc_x=10,
            loc_y=font_size * i,
        )

    for i, result in enumerate(task_results[TaskType.OBJECT_DETECTION], start=1):
        image = draw_object_detection(image, result, names=names)

    for i, result in enumerate(task_results[TaskType.INSTANCE_SEGMENTATION], start=1):
        image = draw_instance_segmentation(image, result, names=names)

    for i, result in enumerate(task_results[TaskType.TEXT_RECOGNITION], start=1):
        image = draw_text_recognition(image, result, loc_x=10, loc_y=10)

    for i, result in enumerate(task_results[TaskType.SEMANTIC_SEGMENTATION], start=1):
        image = draw_semantic_segmentation(image, result, names=names)

    return image



from dataclasses import dataclass
from typing import Generic, TypeVar

from waffle_hub import (
    BaseEnum,
    EvaluateStatus,
    ExportStatus,
    InferenceStatus,
    TrainStatus,
)
from waffle_hub.schema.base_schema import BaseSchema

STATUS_TYPE = TypeVar("STATUS_TYPE", bound=BaseEnum)


@dataclass
class BaseWorkingInfo(Generic[STATUS_TYPE], BaseSchema):
    status: STATUS_TYPE = None
    error_type: str = None
    error_msg: str = None
    step: int = None
    total_step: int = None

    def __setattr__(self, name, value):
        if name == "status":
            if not (value is None or value in self.__class__.__orig_bases__[0].__args__[0]):
                raise ValueError(
                    f"status must be one of {self.__class__.__orig_bases__[0].__args__[0]}"
                )
        self.__dict__[name] = value


@dataclass
class TrainingInfo(BaseWorkingInfo[TrainStatus]):
    pass


@dataclass
class EvaluatingInfo(BaseWorkingInfo[EvaluateStatus]):
    pass


@dataclass
class InferencingInfo(BaseWorkingInfo[InferenceStatus]):
    pass


@dataclass
class ExportingInfo(BaseWorkingInfo[ExportStatus]):
    pass



import fire
from waffle_utils.log import initialize_logger

from waffle_hub.dataset import Dataset
from waffle_hub.utils.base_cli import BaseCLI, cli

initialize_logger("dataset.log", root_level="INFO", console_level="INFO", file_level="DEBUG")


class DatasetInstance(BaseCLI):
    # not required for cli mode
    restrict_method_names = [
        "load",
        "new",
        "add_annotations",
        "add_categories",
        "add_images",
        "add_predictions",
        "create_index",
        "initialize",
        "save_dataset_info",
    ]

    def __init__(self, name: str, root_dir: str = None):
        self.dataset = None
        if name in Dataset.get_dataset_list(root_dir):
            self.dataset = Dataset.load(name, root_dir=root_dir)
        else:
            raise ValueError(f"Dataset {name} does not exist.")

        super().__init__()

    def get_object(self):
        return self.dataset


def main():
    fire.Fire(cli(Dataset, DatasetInstance), serialize=str)


if __name__ == "__main__":
    main()



import warnings
from pathlib import Path
from typing import Union

import cv2
import numpy as np
import torch
from natsort import natsorted
from torchvision import transforms as T
from waffle_utils.file import io
from waffle_utils.image.io import load_image

from waffle_hub.dataset import Dataset
from waffle_hub.schema.data import ImageInfo
from waffle_hub.schema.fields import Annotation, Category, Image

IMAGE_EXTS = [".png", ".jpg", ".jpeg", ".bmp", ".tif", ".tiff"]
VIDEO_EXTS = [".mp4", ".avi", ".mov", ".mkv"]


def get_images(d, recursive: bool = True) -> list[str]:
    exp = "**/*" if recursive else "*"
    image_paths = []
    for ext in IMAGE_EXTS:
        image_paths += list(Path(d).glob(exp.lower() + ext))
        image_paths += list(Path(d).glob(exp.upper() + ext))
    return natsorted(
        list(
            set(
                map(
                    str,
                    image_paths,
                )
            )
        )
    )


def get_videos(d, recursive: bool = True) -> list[str]:
    exp = "**/*" if recursive else "*"
    video_paths = []
    for ext in VIDEO_EXTS:
        video_paths += list(Path(d).glob(exp.lower() + ext))
        video_paths += list(Path(d).glob(exp.upper() + ext))
    return natsorted(
        list(
            set(
                map(
                    str,
                    video_paths,
                )
            )
        )
    )


def resize_image(
    image: np.ndarray, image_size: list[int], letter_box: bool = False
) -> list[np.ndarray, ImageInfo]:
    """Resize Image.

    Args:
        image (np.ndarray): opencv image.
        image_size (list[int]): image [width, height].
        letter_box (bool): letter box.

    Returns:
        list[np.ndarray, ImageInfo]: resized image, image info.
    """

    h, w = image.shape[:2]
    W, H = image_size

    if letter_box:
        h_ratio = H / h
        w_ratio = W / w

        if w_ratio < h_ratio:
            resize_shape = (round(w * w_ratio), round(h * w_ratio))
            total_pad = H - resize_shape[1]
            top = total_pad // 2
            bottom = total_pad - top
            left, right = 0, 0
        else:
            resize_shape = (round(w * h_ratio), round(h * h_ratio))
            total_pad = W - resize_shape[0]
            left = total_pad // 2
            right = total_pad - left
            top, bottom = 0, 0

        resized_image = cv2.resize(image, resize_shape, interpolation=cv2.INTER_LINEAR)
        resized_image = cv2.copyMakeBorder(
            resized_image, top, bottom, left, right, None, value=(114, 114, 114)
        )

    else:
        resize_shape = (W, H)
        left, top = 0, 0
        resized_image = cv2.resize(image, resize_shape, interpolation=cv2.INTER_LINEAR)

    return resized_image, ImageInfo(
        ori_image=cv2.cvtColor(image, cv2.COLOR_BGR2RGB),
        ori_shape=(w, h),
        new_shape=resize_shape,
        input_shape=(W, H),
        pad=(left, top),
    )


def get_image_transform(image_size: Union[int, list[int]], letter_box: bool = False):
    def transform(image: Union[np.ndarray, str]) -> tuple[torch.Tensor, ImageInfo]:
        if isinstance(image, (str, Path)):
            image = load_image(image)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image, image_info = resize_image(image, image_size, letter_box)
        return T.ToTensor()(image), image_info

    return transform


def get_dataset_class(dataset_type: str):
    if dataset_type == "image":
        return ImageDataset
    elif dataset_type == "video":
        return VideoDataset
    elif dataset_type == "dataset":
        return LabeledDataset
    else:
        raise ValueError(f"Unknown dataset type: {dataset_type}")


class BaseDataset:
    def __init__(
        self,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
    ):
        if isinstance(image_size, int):
            image_size = [image_size, image_size]

        if len(image_size) != 2:
            raise ValueError("image_size must be a int or list of int with length 2(width, height).")

        self.image_size = image_size
        self.letter_box = letter_box

        self.transform = get_image_transform(self.image_size, self.letter_box)

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, idx):
        raise NotImplementedError

    def collate_fn(self, batch):
        raise NotImplementedError

    def get_dataloader(self, batch_size: int = 4, num_workers: int = 0):
        return torch.utils.data.DataLoader(
            self,
            batch_size,
            num_workers=num_workers,
            collate_fn=self.collate_fn,
            shuffle=False,
            drop_last=False,
        )


class ImageDataset(BaseDataset):
    def __init__(
        self,
        image_dir: str,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
        recursive: bool = True,
        **kwargs,
    ):
        super().__init__(image_size, letter_box)

        self.image_dir = image_dir
        if Path(self.image_dir).is_file():
            self.image_paths = [self.image_dir]
            self.image_root_dir = Path(self.image_dir).parent
        else:
            self.image_paths = get_images(self.image_dir, recursive=recursive)
            self.image_root_dir = Path(self.image_dir)

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image_path = self.image_paths[idx]

        image_tensor, image_info = self.transform(image_path)
        image_info.image_path = image_path
        image_info.image_rel_path = str(Path(image_path).relative_to(self.image_root_dir))

        return image_tensor, image_info

    def collate_fn(self, batch):
        images, infos = list(zip(*batch))
        return torch.stack(images, dim=0), infos


class LabeledDataset(BaseDataset):
    def __init__(
        self,
        dataset: Dataset,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
        set_name: str = None,
        **kwargs,
    ):
        super().__init__(image_size, letter_box)

        self.dataset = dataset
        self.image_dir = dataset.raw_image_dir
        self.set_name = set_name

        if self.set_name == "train":
            set_file = self.dataset.train_set_file
        elif self.set_name == "val":
            set_file = self.dataset.val_set_file
        elif self.set_name == "test":
            set_file = self.dataset.test_set_file
        else:
            set_file = None

        self.images: list[Image] = self.dataset.get_images(
            io.load_json(set_file) if set_file else None
        )
        self.image_to_annotations: dict[int, list[Annotation]] = {
            image.image_id: self.dataset.get_annotations(image.image_id) for image in self.images
        }

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = self.images[idx]
        image_path = str(self.image_dir / image.file_name)
        annotations: list[Annotation] = self.image_to_annotations[image.image_id]

        image_tensor, image_info = self.transform(image_path)
        image_info.image_path = image_path
        image_info.image_rel_path = image.file_name

        return image_tensor, image_info, annotations

    def collate_fn(self, batch):
        images, infos, annotations = list(zip(*batch))
        return torch.stack(images, dim=0), infos, annotations


class VideoDataset(BaseDataset):
    def __init__(
        self,
        video_path: str,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
        **kwargs,
    ):
        super().__init__(image_size, letter_box)

        self.video_path = video_path
        self.cap = cv2.VideoCapture(self.video_path)
        self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.fps = round(self.cap.get(cv2.CAP_PROP_FPS))

    def __len__(self):
        return self.frame_count

    def __getitem__(self, idx):
        # self.cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
        ret, frame = self.cap.read()
        if not ret:
            raise ValueError(f"Failed to read frame {idx} from video {self.video_path}.")
        image_tensor, image_info = self.transform(frame)
        image_info.image_rel_path = f"{idx}.png"

        return image_tensor, image_info

    def collate_fn(self, batch):
        images, infos = list(zip(*batch))
        return torch.stack(images, dim=0), infos

    def get_dataloader(self, batch_size: int = 1, num_workers: int = 0):
        if batch_size > 1:
            warnings.warn("batch_size > 1 is not supported for video dataset.")
        if num_workers > 0:
            warnings.warn("num_workers > 0 is not supported for video dataset.")
        return super().get_dataloader(batch_size=1, num_workers=0)



import logging
from functools import reduce
from operator import eq
from typing import Union

import torch
from torchmetrics.classification import (
    Accuracy,
    ConfusionMatrix,
    F1Score,
    Precision,
    Recall,
)
from torchmetrics.detection import mean_ap

from waffle_hub import TaskType
from waffle_hub.schema.evaluate import (
    ClassificationMetric,
    InstanceSegmentationMetric,
    ObjectDetectionMetric,
    SemanticSegmentationMetric,
    TextRecognitionMetric,
)
from waffle_hub.schema.fields import Annotation
from waffle_hub.utils.conversion import convert_polygon_to_mask

logger = logging.getLogger(__name__)


def convert_to_torchmetric_format(
    total: list[Annotation], task: TaskType, prediction: bool = False, *args, **kwargs
):

    datas = []
    for annotations in total:

        if task == TaskType.CLASSIFICATION:  # single attribute
            datas.append(annotations[0].category_id - 1)

        elif task == TaskType.OBJECT_DETECTION:
            data = {
                "boxes": [],
                "labels": [],
            }
            if prediction:
                data["scores"] = []

            for annotation in annotations:
                data["boxes"].append(annotation.bbox)
                data["labels"].append(annotation.category_id - 1)
                if prediction:
                    data["scores"].append(annotation.score)

            datas.append(data)

        elif task == TaskType.INSTANCE_SEGMENTATION:
            data = {
                "boxes": [],
                "labels": [],
            }
            if prediction:
                data["scores"] = []

            for annotation in annotations:
                data["boxes"].append(annotation.bbox)
                data["labels"].append(annotation.category_id - 1)
                if prediction:
                    data["scores"].append(annotation.score)

            datas.append(data)

        elif task == TaskType.TEXT_RECOGNITION:
            datas.append(annotations[0].caption)

        elif task == TaskType.SEMANTIC_SEGMENTATION:
            data = {
                "boxes": [],
                "masks": [],
                "labels": [],
            }

            for annotation in annotations:
                data["boxes"].append(annotation.bbox)
                data["masks"].append(
                    (
                        convert_polygon_to_mask(
                            annotation.segmentation, image_size=kwargs["image_size"]
                        )
                    ).tolist()
                )
                data["labels"].append(annotation.category_id - 1)

            datas.append(data)

        else:
            raise NotImplementedError

    if isinstance(datas[0], dict):
        datas = [{k: torch.tensor(v) for k, v in data.items()} for data in datas]
    elif isinstance(datas[0], int):
        datas = torch.tensor(datas)
    elif isinstance(datas[0], str):
        pass
    else:
        raise NotImplementedError

    return datas


def evaluate_classification(
    preds: list[Annotation], labels: list[Annotation], num_classes: int, *args, **kwargs
) -> ClassificationMetric:
    preds = convert_to_torchmetric_format(preds, TaskType.CLASSIFICATION, prediction=True)
    labels = convert_to_torchmetric_format(labels, TaskType.CLASSIFICATION)

    mean_acc = Accuracy(task="multiclass", num_classes=num_classes, average="micro")(preds, labels)
    mean_recall = Recall(task="multiclass", num_classes=num_classes, average="micro")(preds, labels)
    mean_precision = Precision(task="multiclass", num_classes=num_classes, average="micro")(
        preds, labels
    )
    mean_f1_score = F1Score(task="multiclass", num_classes=num_classes, average="micro")(
        preds, labels
    )

    accs = Accuracy(task="multiclass", num_classes=num_classes, average="none")(preds, labels)
    recalls = Recall(task="multiclass", num_classes=num_classes, average="none")(preds, labels)
    precisions = Precision(task="multiclass", num_classes=num_classes, average="none")(preds, labels)
    f1_scores = F1Score(task="multiclass", num_classes=num_classes, average="none")(preds, labels)
    confmats = ConfusionMatrix(task="multiclass", num_classes=num_classes)(preds, labels)

    result = ClassificationMetric(
        accuracy=float(mean_acc),
        recall=float(mean_recall),
        precision=float(mean_precision),
        f1_score=float(mean_f1_score),
        accuracy_per_class=accs.tolist(),
        recall_per_class=recalls.tolist(),
        precision_per_class=precisions.tolist(),
        f1_score_per_class=f1_scores.tolist(),
        confusion_matrix=confmats.tolist(),
    )
    return result


def evaluate_object_detection(
    preds: list[Annotation], labels: list[Annotation], num_classes: int, *args, **kwargs
) -> ObjectDetectionMetric:
    preds = convert_to_torchmetric_format(preds, TaskType.OBJECT_DETECTION, prediction=True)
    labels = convert_to_torchmetric_format(labels, TaskType.OBJECT_DETECTION)

    map_dict = mean_ap.MeanAveragePrecision(
        box_format="xywh",
        iou_type="bbox",
        class_metrics=True,
    )(preds, labels)

    result = ObjectDetectionMetric(
        mAP=float(map_dict["map"]),
        mAP_50=float(map_dict["map_50"]),
        mAP_75=float(map_dict["map_75"]),
        mAP_small=float(map_dict["map_small"]),
        mAP_medium=float(map_dict["map_medium"]),
        mAP_large=float(map_dict["map_large"]),
        mAR_1=float(map_dict["mar_1"]),
        mAR_10=float(map_dict["mar_10"]),
        mAR_100=float(map_dict["mar_100"]),
        mAR_small=float(map_dict["mar_small"]),
        mAR_medium=float(map_dict["map_medium"]),
        mAR_large=float(map_dict["map_large"]),
        mAP_per_class=map_dict["map_per_class"].tolist(),
        mAR_100_per_class=map_dict["mar_100_per_class"].tolist(),
    )
    return result


def evaluate_instance_segmentation(
    preds: list[Annotation], labels: list[Annotation], num_classes: int, *args, **kwargs
) -> InstanceSegmentationMetric:
    preds = convert_to_torchmetric_format(preds, TaskType.INSTANCE_SEGMENTATION, prediction=True)
    labels = convert_to_torchmetric_format(labels, TaskType.INSTANCE_SEGMENTATION)

    map_dict = mean_ap.MeanAveragePrecision(
        box_format="xywh",
        iou_type="bbox",
        class_metrics=True,
    )(preds, labels)

    result = InstanceSegmentationMetric(float(map_dict["map"]))
    return result


def evalute_text_recognition(
    preds: list[Annotation], labels: list[Annotation], num_classes: int, *args, **kwargs
) -> ObjectDetectionMetric:
    preds = convert_to_torchmetric_format(preds, TaskType.TEXT_RECOGNITION, prediction=True)
    labels = convert_to_torchmetric_format(labels, TaskType.TEXT_RECOGNITION)

    correct = reduce(lambda n, pair: n + eq(*pair), zip(preds, labels), 0)
    acc = correct / len(preds)

    result = TextRecognitionMetric(float(acc))
    return result


def evalute_semantic_segmentation(
    preds: list[Annotation],
    labels: list[Annotation],
    num_classes: int,
    image_size: list[int],
    *args,
    **kwargs
) -> SemanticSegmentationMetric:
    preds = convert_to_torchmetric_format(
        preds, TaskType.SEMANTIC_SEGMENTATION, prediction=True, image_size=image_size
    )
    labels = convert_to_torchmetric_format(
        labels, TaskType.SEMANTIC_SEGMENTATION, image_size=image_size
    )
    # TODO: use library

    # mpa
    mean_pixel_accuracy = 0
    for pred, label in zip(preds, labels):
        if pred["masks"].numel() == 0:  # If the object isn't detected
            continue
        
        _mpa = 0
        for label_index, class_id in enumerate(label["labels"]):
            pred_index = torch.where(pred["labels"] == class_id)[0]
            if pred_index.numel() == 0:
                continue

            _mpa += torch.sum(pred["masks"][pred_index] == label["masks"][label_index]) / torch.numel(label["masks"][label_index])
        mean_pixel_accuracy += _mpa / len(label["labels"])
    mean_pixel_accuracy /= len(labels)

    # iou
    iou = 0
    for pred, label in zip(preds, labels):
        if pred["masks"].numel() == 0:  # If the object isn't detected
            continue
        
        _iou = 0
        for label_index, class_id in enumerate(label["labels"]):
            pred_index = torch.where(pred["labels"] == class_id)[0]
            if pred_index.numel() == 0:
                continue

            label_mask = (label["masks"][label_index] == 255)
            pred_mask = (pred["masks"][pred_index] == 255)

            intersection = torch.sum(pred_mask & label_mask)
            union = torch.sum(pred_mask) + torch.sum(label_mask) - intersection

            if union == 0:
                continue

            _iou += (intersection / union)
        iou += _iou / len(label["labels"])
    iou /= len(labels)

    result = SemanticSegmentationMetric(float(mean_pixel_accuracy), float(iou))
    return result


def evaluate_function(
    preds: list[Annotation],
    labels: list[Annotation],
    task: str,
    num_classes: int = None,
    *args,
    **kwargs
) -> Union[
    ClassificationMetric,
    ObjectDetectionMetric,
    InstanceSegmentationMetric,
    TextRecognitionMetric,
    SemanticSegmentationMetric,
]:
    if task == TaskType.CLASSIFICATION:
        return evaluate_classification(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.OBJECT_DETECTION:
        return evaluate_object_detection(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.INSTANCE_SEGMENTATION:
        return evaluate_instance_segmentation(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.TEXT_RECOGNITION:
        return evalute_text_recognition(preds, labels, num_classes, *args, **kwargs)
    elif task == TaskType.SEMANTIC_SEGMENTATION:
        return evalute_semantic_segmentation(preds, labels, num_classes, *args, **kwargs)
    else:
        raise NotImplementedError



__version__ = "0.2.16"

import enum
import signal
from collections import OrderedDict


class CustomEnumMeta(enum.EnumMeta):
    def __contains__(cls, item):
        if isinstance(item, str):
            return item.upper() in cls._member_names_
        return super().__contains__(item)

    def __upper__(self):
        return self.name.upper()


class BaseEnum(enum.Enum, metaclass=CustomEnumMeta):
    """Base class for Enum

    Example:
        >>> class Color(BaseEnum):
        >>>     RED = 1
        >>>     GREEN = 2
        >>>     BLUE = 3
        >>> Color.RED == "red"
        True
        >>> Color.RED == "RED"
        True
        >>> "red" in DataType
        True
        >>> "RED" in DataType
        True
    """

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name.upper() == other.upper()
        return super().__eq__(other)

    def __ne__(self, other):
        if isinstance(other, str):
            return self.name.upper() != other.upper()
        return super().__ne__(other)

    def __hash__(self):
        return hash(self.name.upper())

    def __str__(self):
        return self.name.upper()

    def __repr__(self):
        return self.name.upper()


class DataType(BaseEnum):
    # TODO: map to same value

    YOLO = enum.auto()
    ULTRALYTICS = enum.auto()

    COCO = enum.auto()

    AUTOCARE_DLT = enum.auto()

    TRANSFORMERS = enum.auto()


class TaskType(BaseEnum):
    CLASSIFICATION = enum.auto()
    OBJECT_DETECTION = enum.auto()
    SEMANTIC_SEGMENTATION = enum.auto()
    INSTANCE_SEGMENTATION = enum.auto()
    KEYPOINT_DETECTION = enum.auto()
    TEXT_RECOGNITION = enum.auto()
    REGRESSION = enum.auto()


class SplitMethod(BaseEnum):
    RANDOM = enum.auto()
    STRATIFIED = enum.auto()


# for changeable status
class TrainStatus(BaseEnum):
    INIT = enum.auto()
    RUNNING = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()
    STOPPED = enum.auto()


class EvaluateStatus(BaseEnum):
    INIT = enum.auto()
    RUNNING = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()
    STOPPED = enum.auto()


class InferenceStatus(BaseEnum):
    INIT = enum.auto()
    RUNNING = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()
    STOPPED = enum.auto()


class ExportStatus(BaseEnum):
    INIT = enum.auto()
    RUNNING = enum.auto()
    SUCCESS = enum.auto()
    FAILED = enum.auto()
    STOPPED = enum.auto()


EXPORT_MAP = OrderedDict(
    {
        DataType.YOLO: "ULTRALYTICS",
        DataType.ULTRALYTICS: "ULTRALYTICS",
        DataType.COCO: "COCO",
        DataType.AUTOCARE_DLT: "AUTOCARE_DLT",
        DataType.TRANSFORMERS: "TRANSFORMERS",
    }
)


BACKEND_MAP = OrderedDict(
    {
        DataType.ULTRALYTICS: {
            "import_path": "waffle_hub.hub.adapter.ultralytics",
            "class_name": "UltralyticsHub",
        },
        DataType.AUTOCARE_DLT: {
            "import_path": "waffle_hub.hub.adapter.autocare_dlt",
            "class_name": "AutocareDLTHub",
        },
        DataType.TRANSFORMERS: {
            "import_path": "waffle_hub.hub.adapter.transformers",
            "class_name": "TransformersHub",
        },
    }
)


for key in list(EXPORT_MAP.keys()):
    EXPORT_MAP[str(key).lower()] = EXPORT_MAP[key]
    EXPORT_MAP[str(key).upper()] = EXPORT_MAP[key]


for key in list(BACKEND_MAP.keys()):
    BACKEND_MAP[str(key).lower()] = BACKEND_MAP[key]
    BACKEND_MAP[str(key).upper()] = BACKEND_MAP[key]


# except handler for SIGINT, SIGTERM, SIGCHILD
def sigint_handler(signum, frame):
    raise KeyboardInterrupt


def sigterm_handler(signum, frame):
    raise SystemExit


signal.signal(signal.SIGINT, sigint_handler)
signal.signal(signal.SIGTERM, sigterm_handler)



import subprocess
from pathlib import Path

import pytest

from waffle_hub import TaskType
from waffle_hub.dataset import Dataset


def run_cli(cmd):
    ret = subprocess.run(cmd, check=True, shell=True)
    return ret


# hub test
def _new(hub_name: str, tmpdir: Path, task: TaskType):
    cmd = f"python -m waffle_hub.hub.cli new \
        --backend ultralytics \
        --root-dir {tmpdir} \
        --name {hub_name} \
        --task {task} \
        --model-type yolov8 \
        --model-size n \
        --categories [1,2] \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name).exists()


def _train(hub_name: str, dataset: Dataset, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli train \
        --root-dir {tmpdir} \
        --name {hub_name} \
        --dataset {dataset.name} \
        --dataset_root_dir {dataset.root_dir} \
        --epochs 1 \
        --batch-size 4 \
        --image-size 64 \
        --device cpu \
        --workers 0 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / "artifacts").exists()
    assert (tmpdir / hub_name / "working_infos" / "training_info.json").exists()


def _delete_artifact(hub_name: str, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli delete_artifact \
        --name  {hub_name} \
        --root-dir {tmpdir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert not (tmpdir / hub_name / "artifacts").exists()


def _train_advance_params(hub_name: str, dataset: Dataset, tmpdir: Path):
    cmd = (
        f"python -m waffle_hub.hub.cli train \
        --root-dir {tmpdir} \
        --name {hub_name} \
        --dataset {dataset.name} \
        --dataset_root_dir {dataset.root_dir} \
        --epochs 1 \
        --batch-size 4 \
        --image-size 64 \
        --device cpu \
        --workers 0 \
    "
        + ' --advance_params "{box: 3}"'
    )
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / "artifacts").exists()
    assert (tmpdir / hub_name / "working_infos" / "training_info.json").exists()


def _inference(hub_name: str, dataset: Dataset, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli inference \
        --root-dir {tmpdir} \
        --name {hub_name} \
        --source {dataset.raw_image_dir} \
        --confidence-threshold 0.25 \
        --device cpu \
        --workers 0 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / "inferences").exists()
    assert (tmpdir / hub_name / "working_infos" / "inferencing_info.json").exists()


def _evaluate(hub_name: str, dataset: Dataset, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli evaluate \
        --name {hub_name} \
        --root-dir {tmpdir} \
        --dataset {dataset.name} \
        --dataset_root_dir {dataset.root_dir} \
        --set-name test \
        --device cpu \
        --workers 0 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / "evaluate.json").exists()
    assert (tmpdir / hub_name / "working_infos" / "evaluating_info.json").exists()


def _export_onnx(hub_name: str, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli export_onnx \
        --name {hub_name} \
        --root-dir {tmpdir} \
        --device cpu \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / "weights" / "model.onnx").exists()
    assert (tmpdir / hub_name / "working_infos" / "exporting_onnx_info.json").exists()


def _export_waffle(hub_name: str, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli export_waffle \
        --name {hub_name} \
        --root-dir {tmpdir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name / f"{hub_name}.waffle").exists()


def _from_waffle_file(hub_name: str, dataset: Dataset, tmpdir: Path):
    cmd = f'python -m waffle_hub.hub.cli from_waffle_file \
        --name from_waffle_file_test \
        --waffle_file {tmpdir / hub_name / f"{hub_name}.waffle"} \
        --root-dir {tmpdir} \
    '
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / hub_name).exists()

    # test_from_waffle_file_inference
    cmd = f"python -m waffle_hub.hub.cli inference \
        --root-dir {tmpdir} \
        --name from_waffle_file_test \
        --source {dataset.raw_image_dir} \
        --confidence-threshold 0.25 \
        --device cpu \
        --workers 0 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / "from_waffle_file_test" / "inferences").exists()


def _delete_hub(hub_name: str, tmpdir: Path):
    cmd = f"python -m waffle_hub.hub.cli delete_hub \
        --name  {hub_name} \
        --root-dir {tmpdir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert not (tmpdir / hub_name).exists()


def test_hub(tmpdir: Path, object_detection_dataset: Dataset):
    hub_name = "test_hub"
    dataset = object_detection_dataset
    _new(hub_name, tmpdir, TaskType.OBJECT_DETECTION)
    _train(hub_name, dataset, tmpdir)
    _inference(hub_name, dataset, tmpdir)
    _evaluate(hub_name, dataset, tmpdir)
    _export_onnx(hub_name, tmpdir)
    _export_waffle(hub_name, tmpdir)
    _from_waffle_file(hub_name, dataset, tmpdir)
    _delete_hub(hub_name, tmpdir)


# dataset
def _split(dataset_name: str, tmpdir: Path):
    cmd = f"python -m waffle_hub.dataset.cli split \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --train-ratio 0.8 \
        --val-ratio 0.1 \
        --test-ratio 0.1 \
        --method random \
        --seed 42 \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name / "sets" / "train.json").exists()


def _export(dataset_name: str, tmpdir: Path):
    for data_type in ["ULTRALYTICS", "COCO"]:
        cmd = f"python -m waffle_hub.dataset.cli export \
            --data-type {data_type} \
            --name {dataset_name} \
            --root-dir {tmpdir} \
        "
        ret = run_cli(cmd)
        assert ret.returncode == 0
        assert (tmpdir / dataset_name / "exports" / data_type).exists()


@pytest.mark.parametrize("task", [TaskType.CLASSIFICATION, TaskType.OBJECT_DETECTION])
def test_dataset_coco(coco_path: Path, tmpdir: Path, task: TaskType):
    dataset_name = "from_coco"
    cmd = f"python -m waffle_hub.dataset.cli from_coco \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --coco-file {coco_path / 'coco.json'} \
        --coco-root-dir {coco_path / 'images'} \
        --task {task} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name).exists()
    _split(dataset_name, tmpdir)
    _export(dataset_name, tmpdir)


def test_dataset_yolo_cls(yolo_classification_path: Path, tmpdir: Path):
    dataset_name = "from_yolo"
    cmd = f"python -m waffle_hub.dataset.cli from_yolo \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --task classification \
        --yolo-root-dir {yolo_classification_path} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name).exists()
    _split(dataset_name, tmpdir)
    _export(dataset_name, tmpdir)


def test_dataset_yolo_obj(yolo_object_detection_path: Path, tmpdir: Path):
    dataset_name = "from_yolo"
    cmd = f"python -m waffle_hub.dataset.cli from_yolo \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --task object_detection \
        --yolo-root-dir {yolo_object_detection_path} \
        --yaml-path {yolo_object_detection_path / 'data.yaml'} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name).exists()
    _split(dataset_name, tmpdir)
    _export(dataset_name, tmpdir)


def test_dataset_transformer_cls(transformers_classification_path: Path, tmpdir: Path):
    dataset_name = "from_hf"
    cmd = f"python -m waffle_hub.dataset.cli from_transformers \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --task classification \
        --dataset-dir {transformers_classification_path} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name).exists()
    _split(dataset_name, tmpdir)
    _export(dataset_name, tmpdir)


def test_dataset_transformer_obj(transformers_detection_path: Path, tmpdir: Path):
    dataset_name = "from_hf"
    cmd = f"python -m waffle_hub.dataset.cli from_transformers \
        --name {dataset_name} \
        --root-dir {tmpdir} \
        --task object_detection \
        --dataset-dir {transformers_detection_path} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / dataset_name).exists()
    _split(dataset_name, tmpdir)
    _export(dataset_name, tmpdir)


def test_dataset_clone(object_detection_dataset: Dataset, tmpdir: Path):
    dataset = object_detection_dataset
    cmd = f"python -m waffle_hub.dataset.cli clone \
        --src-name {dataset.name} \
        --src-root-dir {dataset.root_dir} \
        --name clone \
        --root-dir {tmpdir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / "clone").exists()


def test_dataset_delete(object_detection_dataset: Dataset):
    dataset = object_detection_dataset
    dataset_path = dataset.dataset_dir
    cmd = f"python -m waffle_hub.dataset.cli delete \
        --name {dataset.name} \
        --root-dir {dataset.root_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert not (dataset_path).exists()


def test_dataset_get_split_ids(object_detection_dataset: Dataset):
    dataset = object_detection_dataset
    cmd = f"python -m waffle_hub.dataset.cli get_split_ids \
        --name {dataset.name} \
        --root-dir {dataset.root_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0


def test_dataset_merge(coco_path: Path, yolo_object_detection_path: Path, tmpdir: Path):
    dataset1_name = "coco"
    dataset2_name = "yolo"
    Dataset.from_coco(
        name=dataset1_name,
        root_dir=tmpdir,
        coco_file=coco_path / "coco.json",
        coco_root_dir=coco_path / "images",
        task=TaskType.OBJECT_DETECTION,
    )
    Dataset.from_yolo(
        name=dataset2_name,
        root_dir=tmpdir,
        task=TaskType.OBJECT_DETECTION,
        yolo_root_dir=yolo_object_detection_path,
        yaml_path=yolo_object_detection_path / "data.yaml",
    )
    cmd = f"python -m waffle_hub.dataset.cli merge \
        --name merge \
        --root-dir {tmpdir} \
        --src-names [{dataset1_name},{dataset2_name}] \
        --src-root-dirs {tmpdir} \
        --task object_detection \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / "merge").exists()


@pytest.mark.parametrize("task", [TaskType.CLASSIFICATION, TaskType.OBJECT_DETECTION])
def test_dataset_sample(tmpdir: Path, task: TaskType):
    cmd = f"python -m waffle_hub.dataset.cli sample \
        --name sample \
        --root-dir {tmpdir} \
        --task {task} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0
    assert (tmpdir / "sample").exists()


def test_dataset_get_fields(object_detection_dataset: Dataset):
    dataset = object_detection_dataset
    # image
    cmd = f"python -m waffle_hub.dataset.cli get_images \
        --name {dataset.name} \
        --root-dir {dataset.root_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0

    # annotation
    cmd = f"python -m waffle_hub.dataset.cli get_annotations \
        --name {dataset.name} \
        --root-dir {dataset.root_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0

    # category
    cmd = f"python -m waffle_hub.dataset.cli get_categories \
        --name {dataset.name} \
        --root-dir {dataset.root_dir} \
    "
    ret = run_cli(cmd)
    assert ret.returncode == 0



"""
Ultralytics Hub
See Hub documentation for more details about usage.
"""

import logging
import warnings
from pathlib import Path
from typing import Union

import torch
import ultralytics
from torchvision import transforms as T
from ultralytics import YOLO
from ultralytics.yolo.utils import DEFAULT_CFG as YOLO_DEFAULT_ADVANCE_PARAMS
from waffle_utils.file import io

from waffle_hub import DataType, TaskType
from waffle_hub.hub import Hub
from waffle_hub.hub.model.wrapper import ModelWrapper
from waffle_hub.schema.configs import TrainConfig
from waffle_hub.utils.process import run_python_file
from waffle_hub.utils.working_info_logger import TrainingInfoLogger

from .config import DEFAULT_PARAMS, MODEL_TYPES, TASK_MAP

logger = logging.getLogger(__name__)


class UltralyticsHub(Hub):
    BACKEND_NAME = DataType.ULTRALYTICS
    MODEL_TYPES = MODEL_TYPES
    MULTI_GPU_TRAIN = True
    DEFAULT_PARAMS = DEFAULT_PARAMS
    DEFAULT_ADVANCE_PARAMS = dict(YOLO_DEFAULT_ADVANCE_PARAMS)

    TASK_MAP = TASK_MAP

    def __init__(
        self,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        backend: str = None,
        version: str = None,
        *args,
        **kwargs,
    ):
        if backend is not None and UltralyticsHub.BACKEND_NAME != backend:
            raise ValueError(
                f"Backend {backend} is not supported. Please use {UltralyticsHub.BACKEND_NAME}"
            )

        if version is not None and ultralytics.__version__ != version:
            warnings.warn(
                f"You've loaded the Hub created with ultralytics=={version}, \n"
                + f"but the installed version is {ultralytics.__version__}."
            )

        super().__init__(
            name=name,
            backend=UltralyticsHub.BACKEND_NAME,
            version=ultralytics.__version__,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

        self.backend_task_name = self.TASK_MAP[self.task]

    def get_default_advance_train_params(
        cls, task: str = None, model_type: str = None, model_size: str = None
    ) -> dict:
        return cls.DEFAULT_ADVANCE_PARAMS

    @classmethod
    def new(
        cls,
        name: str,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        categories: Union[list[dict], list] = None,
        root_dir: str = None,
        *args,
        **kwargs,
    ):
        """Create Ultralytics Hub.

        Args:
            name (str): Hub name
            task (str, optional): Task Name. See UltralyticsHub.TASKS. Defaults to None.
            model_type (str, optional): Model Type. See UltralyticsHub.MODEL_TYPES. Defaults to None.
            model_size (str, optional): Model Size. See UltralyticsHub.MODEL_SIZES. Defaults to None.
            categories (Union[list[dict], list]): class dictionary or list. [{"supercategory": "name"}, ] or ["name",].
            root_dir (str, optional): Root directory of hub repository. Defaults to None.
        """

        warnings.warn("UltralyticsHub.new() is deprecated. Please use Hub.new() instead.")

        return cls(
            name=name,
            task=task,
            model_type=model_type,
            model_size=model_size,
            categories=categories,
            root_dir=root_dir,
        )

    # Hub Utils
    def get_preprocess(self, *args, **kwargs):

        if self.task == TaskType.CLASSIFICATION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.OBJECT_DETECTION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        elif self.task == TaskType.INSTANCE_SEGMENTATION:
            normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

            def preprocess(x, *args, **kwargs):
                return normalize(x)

        else:
            raise NotImplementedError(f"Task {self.task} is not implemented.")

        return preprocess

    def get_postprocess(self, *args, **kwargs):

        id_mapper: list[int] = kwargs.get("id_mapper", [i for i in range(len(self.categories))])

        if self.task == TaskType.CLASSIFICATION:

            def inner(x: torch.Tensor, *args, **kwargs):
                return [x[:, id_mapper]]

        elif self.task == TaskType.OBJECT_DETECTION:

            def inner(x: torch.Tensor, image_size: tuple[int, int], *args, **kwargs):
                x = x[0]  # x[0]: prediction, x[1]: TODO: what is this...?
                x = x.transpose(1, 2)

                cxcywh = x[:, :, :4]
                cx, cy, w, h = torch.unbind(cxcywh, dim=-1)
                x1 = cx - w / 2
                y1 = cy - h / 2
                x2 = cx + w / 2
                y2 = cy + h / 2
                xyxy = torch.stack([x1, y1, x2, y2], dim=-1)

                xyxy[:, :, ::2] /= image_size[0]
                xyxy[:, :, 1::2] /= image_size[1]
                probs = x[:, :, 4:][:, :, id_mapper]
                confidences, class_ids = torch.max(probs, dim=-1)

                return xyxy, confidences, class_ids

        elif self.task == TaskType.INSTANCE_SEGMENTATION:
            num_category = len(self.categories)

            def inner(x: torch.Tensor, image_size: tuple[int, int], *args, **kwargs):
                preds = x[0]  # x[0]: prediction, x[1]: TODO: what is this...?
                preds = preds.transpose(1, 2)

                probs = preds[:, :, 4 : 4 + num_category]
                confidences, class_ids = torch.max(probs, dim=-1)
                _, indicies = torch.topk(
                    confidences, k=min(100, confidences.shape[-1]), dim=-1, largest=True
                )  # TODO: make k configurable

                preds = torch.gather(preds, 1, indicies.unsqueeze(-1).repeat(1, 1, preds.shape[-1]))
                confidences = torch.gather(confidences, 1, indicies)
                class_ids = torch.gather(class_ids, 1, indicies)

                cxcywh = preds[:, :, :4]
                cx, cy, w, h = torch.unbind(cxcywh, dim=-1)
                x1 = cx - w / 2
                y1 = cy - h / 2
                x2 = cx + w / 2
                y2 = cy + h / 2
                xyxy = torch.stack([x1, y1, x2, y2], dim=-1)

                xyxy[:, :, ::2] /= image_size[0]
                xyxy[:, :, 1::2] /= image_size[1]

                # [batch, mask_dim, mask_height, mask_width] -> [batch, num_mask, mask_height * mask_width]
                protos = x[1][-1] if len(x[1]) == 3 else x[1]
                mask_size = (protos.shape[-2], protos.shape[-1])
                protos = protos.view(
                    protos.shape[0], -1, protos.shape[-2] * protos.shape[-1]
                ).float()

                masks = preds[:, :, 4 + num_category :]
                masks = torch.bmm(masks, protos).sigmoid().view(masks.shape[0], -1, *mask_size)

                return xyxy, confidences, class_ids, masks

        else:
            raise NotImplementedError(f"Task {self.task} is not implemented.")

        return inner

    def get_metrics(self) -> list[list[dict]]:
        # read csv file
        # epoch,         train/box_loss,         train/cls_loss,         train/dfl_loss,   metrics/precision(B),      metrics/recall(B),       metrics/mAP50(B),    metrics/mAP50-95(B),           val/box_loss,           val/cls_loss,           val/dfl_loss,                 lr/pg0,                 lr/pg1,                 lr/pg2
        #              0,                 2.0349,                 4.4739,                  1.214,                0.75962,                0.34442,                0.22858,                0.18438,                0.61141,                 2.8409,                0.78766,                 0.0919,                 0.0009,                 0.0009
        #              1,                 1.7328,                 4.0078,                 1.1147,                0.12267,                0.40909,                0.20891,                0.15007,                0.75082,                 2.7157,                0.79723,               0.082862,              0.0018624,              0.0018624
        # and parse it to list[dict]
        # [[{"tag": train/box_loss, "value": 0.0}, {"tag": train/box_loss, "value": 1.7328}, ...], ...]
        # and return it

        csv_path = self.artifact_dir / "results.csv"

        if not csv_path.exists():
            return []

        with open(csv_path) as f:
            lines = f.readlines()

        header = lines[0].strip().split(",")
        metrics = []
        for line in lines[1:]:
            values = line.strip().split(",")
            metric = []
            for i, value in enumerate(values):
                metric.append(
                    {
                        "tag": header[i].strip(),
                        "value": float(value),
                    }
                )
            metrics.append(metric)

        return metrics

    # Train Hook
    def on_train_start(self, cfg: TrainConfig):

        # set data
        cfg.dataset_path: Path = Path(cfg.dataset_path)
        if self.backend_task_name in ["detect", "segment"]:
            if cfg.dataset_path.suffix not in [".yml", ".yaml"]:
                yaml_files = list(cfg.dataset_path.glob("*.yaml")) + list(
                    cfg.dataset_path.glob("*.yml")
                )
                if len(yaml_files) != 1:
                    raise FileNotFoundError(f"Ambiguous data file. Detected files: {yaml_files}")
                cfg.dataset_path = Path(yaml_files[0]).absolute()
            else:
                cfg.dataset_path = cfg.dataset_path.absolute()
        elif self.backend_task_name == "classify":
            if not cfg.dataset_path.is_dir():
                raise ValueError(
                    f"Classification dataset should be directory. Not {cfg.dataset_path}"
                )
            cfg.dataset_path = cfg.dataset_path.absolute()
        cfg.dataset_path = str(cfg.dataset_path)

        # pretrained model
        cfg.pretrained_model = (
            cfg.pretrained_model
            if cfg.pretrained_model
            else self.MODEL_TYPES[self.task][self.model_type][self.model_size]
        )

        # other
        if self.task in [TaskType.OBJECT_DETECTION, TaskType.INSTANCE_SEGMENTATION]:
            cfg.letter_box = True  # TODO: hard coding
            logger.warning(
                "letter_box False is not supported for Object Detection and Segmentation."
            )

    def training(self, cfg: TrainConfig, info_logger: TrainingInfoLogger):
        params = {
            "data": str(cfg.dataset_path).replace("\\", "/"),
            "epochs": cfg.epochs,
            "batch": cfg.batch_size,
            "imgsz": cfg.image_size,
            "lr0": cfg.learning_rate,
            "lrf": cfg.learning_rate,
            "rect": False,  # TODO: hard coding for mosaic
            "device": cfg.device,
            "workers": cfg.workers,
            "seed": cfg.seed,
            "verbose": cfg.verbose,
            "project": str(self.hub_dir),
            "name": str(self.ARTIFACT_DIR),
        }
        params.update(cfg.advance_params)

        code = f"""if __name__ == "__main__":
        from ultralytics import YOLO
        import os
        os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"
        try:
            model = YOLO("{cfg.pretrained_model}", task="{self.backend_task_name}")
            model.train(
                **{params}
            )
        except Exception as e:
            print(e)
            raise e
        """

        script_file = str((self.hub_dir / "train.py").absolute())
        with open(script_file, "w") as f:
            f.write(code)

        run_python_file(script_file)

    def on_train_end(self, cfg: TrainConfig):
        best_ckpt_file = self.artifact_dir / "weights" / "best.pt"
        last_ckpt_file = self.artifact_dir / "weights" / "last.pt"

        if best_ckpt_file.exists():
            io.copy_file(best_ckpt_file, self.best_ckpt_file, create_directory=True)
        if last_ckpt_file.exists():
            io.copy_file(last_ckpt_file, self.last_ckpt_file, create_directory=True)

        metrics = self.get_metrics()
        if metrics:
            io.save_json(metrics, self.metric_file, create_directory=True)

    # Inference Hook
    def get_model(self):
        """Get model.
        Returns:
            ModelWrapper: Model wrapper
        """
        self.check_train_sanity()

        # get model
        model = YOLO(self.best_ckpt_file).model.eval()

        # get adapt functions
        names: list[str] = list(map(lambda x: x["name"], self.categories))
        yolo_names: dict[int, str] = model.names
        if len(names) != len(yolo_names):
            raise ValueError(
                f"Number of categories is not matched. hub: {len(names)} != ultralytics: {len(yolo_names)}"
            )

        id_mapper: list[int] = [i for i in range(len(yolo_names))]

        yolo_names_inv = {v: k for k, v in yolo_names.items()}
        for i, name in enumerate(names):
            id_mapper[i] = yolo_names_inv[name]

        preprocess = self.get_preprocess()
        postprocess = self.get_postprocess(id_mapper=id_mapper)

        # wrap model
        model = ModelWrapper(
            model=model,
            preprocess=preprocess,
            postprocess=postprocess,
        )

        return model



from pathlib import Path

from waffle_hub import EvaluateStatus, ExportStatus, InferenceStatus, TrainStatus
from waffle_hub.schema.working_info import (
    STATUS_TYPE,
    BaseWorkingInfo,
    EvaluatingInfo,
    ExportingInfo,
    InferencingInfo,
    TrainingInfo,
)


class RunningInfoLogger:
    working_info: BaseWorkingInfo = None

    def __init__(self, working_info: BaseWorkingInfo, save_path: Path):
        self.working_info = working_info
        if self.working_info is None:
            raise NotImplementedError("working_info must be not None")
        self.save_path = save_path

    def save(self):
        self.working_info.save_json(save_path=self.save_path)

    def set_status(self, status: STATUS_TYPE):
        self.working_info.status = status
        self.save()

    def set_error(self, e: Exception = None):
        self.working_info.error_type = e.__class__.__name__
        self.working_info.error_msg = e
        self.save()

    def clear_error(self):
        self.working_info.error_type = None
        self.working_info.error_msg = None
        self.save()

    def clear_step(self):
        self.working_info.step = 0
        self.working_info.total_step = 0
        self.save()

    def set_total_step(self, total_step: int):
        self.working_info.total_step = total_step
        self.working_info.step = 0
        self.save()

    def set_current_step(self, step: int):
        self.working_info.step = step
        self.save()


class TrainingInfoLogger(RunningInfoLogger):
    def __init__(self, save_path: Path):
        super().__init__(TrainingInfo(), save_path)
        self.set_init()

    def set_init(self):
        self.clear_error()
        self.clear_step()
        self.set_status(TrainStatus.INIT)

    def set_failed(self, e):
        self.set_error(e)
        self.set_status(TrainStatus.FAILED)

    def set_success(self):
        self.clear_error()
        self.working_info.step = self.working_info.total_step
        self.set_status(TrainStatus.SUCCESS)

    def set_running(self):
        self.set_status(TrainStatus.RUNNING)

    def set_stopped(self, e):
        self.set_error(e)
        self.set_status(TrainStatus.STOPPED)


class EvaluatingInfoLogger(RunningInfoLogger):
    def __init__(self, save_path: Path):
        super().__init__(EvaluatingInfo(), save_path)
        self.set_init()

    def set_init(self):
        self.clear_error()
        self.clear_step()
        self.set_status(EvaluateStatus.INIT)

    def set_failed(self, e):
        self.set_error(e)
        self.set_status(EvaluateStatus.FAILED)

    def set_success(self):
        self.clear_error()
        self.working_info.step = self.working_info.total_step
        self.set_status(EvaluateStatus.SUCCESS)

    def set_running(self):
        self.set_status(EvaluateStatus.RUNNING)

    def set_stopped(self, e):
        self.set_error(e)
        self.set_status(EvaluateStatus.STOPPED)


class InferencingInfoLogger(RunningInfoLogger):
    def __init__(self, save_path: Path):
        super().__init__(InferencingInfo(), save_path)
        self.set_init()

    def set_init(self):
        self.clear_error()
        self.clear_step()
        self.set_status(InferenceStatus.INIT)

    def set_failed(self, e):
        self.set_error(e)
        self.set_status(InferenceStatus.FAILED)

    def set_success(self):
        self.clear_error()
        self.working_info.step = self.working_info.total_step
        self.set_status(InferenceStatus.SUCCESS)

    def set_running(self):
        self.set_status(InferenceStatus.RUNNING)

    def set_stopped(self, e):
        self.set_error(e)
        self.set_status(InferenceStatus.STOPPED)


class ExportingInfoLogger(RunningInfoLogger):
    def __init__(self, save_path: Path):
        super().__init__(ExportingInfo(), save_path)
        self.set_init()

    def set_init(self):
        self.clear_error()
        self.clear_step()
        self.set_status(ExportStatus.INIT)

    def set_failed(self, e):
        self.set_error(e)
        self.set_status(ExportStatus.FAILED)

    def set_success(self):
        self.clear_error()
        self.working_info.step = self.working_info.total_step
        self.set_status(ExportStatus.SUCCESS)

    def set_running(self):
        self.set_status(ExportStatus.RUNNING)

    def set_stopped(self, e):
        self.set_error(e)
        self.set_status(ExportStatus.STOPPED)
